//
//  PLSAppDelegate.m
//  Pulse
//
//  Created by Matt Attaway on 1/14/14.
//  Copyright (c) 2014 Zen of the Monkey. All rights reserved.
//

#import "PLSAppDelegate.h"
#import "PLSConnectionController.h"
#import "PLSLoginController.h"


@interface PLSAppDelegate()
    @property (nonatomic,strong) NSStatusItem* statusItem;
    @property (nonatomic,strong) NSMenu* statusMenu;
    @property (nonatomic,strong) PLSConnectionController* connectionsController;
    @property (nonatomic,strong) PLSLoginController* loginController;
@end


@implementation PLSAppDelegate


- (void)loadConnections
{
    NSUserDefaults* prefs = [NSUserDefaults standardUserDefaults];
    NSArray* connections = [prefs arrayForKey:@"connections"];
    self.overseers = [[NSMutableArray alloc] init];
    
    for(NSDictionary* connection in connections)
    {
        PLSOverseer* os = [[PLSOverseer alloc] initWithDictionary:connection];
        
        [os startTracking];
        os.delegate = self;
        [self.overseers addObject:os];
    }
}


- (void)buildMenu
{
    self.statusMenu = [[NSMenu alloc] initWithTitle:@""];
    [self.statusMenu setAutoenablesItems:NO];
    
    [self.statusMenu addItemWithTitle:@"Connections" action:@selector(showConnections:)   keyEquivalent:@""];
    [self.statusMenu addItemWithTitle:@"Login" action:@selector(findConnectionNeedingLogin:) keyEquivalent:@""];
    [self.statusMenu addItemWithTitle:@"Submit" action:@selector(launchSubmit:) keyEquivalent:@""];
    [self.statusMenu addItemWithTitle:@"?????" action:nil  keyEquivalent:@""];
    [self.statusMenu addItemWithTitle:@"Profit" action:nil keyEquivalent:@""];
    
    [self.statusMenu addItem:[NSMenuItem separatorItem]];
    
    NSMenuItem* tItem = nil;
    tItem = [self.statusMenu addItemWithTitle:@"Quit" action:@selector(terminate:) keyEquivalent:@"q"];
    [tItem setKeyEquivalentModifierMask:NSCommandKeyMask];
    
    NSStatusBar* statusBar = [NSStatusBar systemStatusBar];
    self.statusItem = [statusBar statusItemWithLength:NSVariableStatusItemLength];
    [self.statusItem setImage:[NSImage imageNamed:@"perforce.png"]];
    [self.statusItem setToolTip:@"Pulse 0.0\nUp-to-date (most likely)"];
    [self.statusItem setHighlightMode:YES];
    [self.statusItem setMenu:self.statusMenu];
    
    // disable the login menu
    [[self.statusMenu itemWithTitle:@"Login"] setEnabled:false];
}


- (void)applicationDidFinishLaunching:(NSNotification*)aNotification
{
    [self loadConnections];
    [self buildMenu];
    
    if([self.overseers count] == 0)
    {
        [self showConnections:self];
    }
}


- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
    NSMutableArray* osSettings = [[NSMutableArray alloc] init];
    
    for(PLSOverseer* os in self.overseers)
    {
        [osSettings addObject:[os getSettings]];
    }
    
    NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
    [defaults setObject:osSettings forKey:@"connections"];
    [defaults synchronize];

    return NSTerminateNow;
}


// show our very classy connection management dialog
- (IBAction)showConnections:(id)sender
{
    if(!self.connectionsController)
    {
        self.connectionsController = [[PLSConnectionController alloc] initWithWindowNibName:@"Connections"];
        self.connectionsController.connections = self.overseers;
        
        // notify self when the connection dialog closes so we can set ourselves as the overseer delegate on new connections
        [[NSNotificationCenter defaultCenter] addObserver:self
                                                 selector:@selector(connectionWindowClosed:)
                                                     name:NSWindowWillCloseNotification
                                                   object:[self.connectionsController window]];
    }
    
    [NSApp activateIgnoringOtherApps:YES];
    [[self.connectionsController window] makeKeyAndOrderFront:self];
    [self.connectionsController showWindow:self];
}


// show the login dialog and use the password to log the user in
- (IBAction)showLogin:(id)sender overseer:(PLSOverseer*)os
{
    // we always create a new login contorller because we need to customize it for each appearance
    self.loginController = [[PLSLoginController alloc] initWithWindowNibName:@"Login"];
    self.loginController.os = os;
    
    [NSApp activateIgnoringOtherApps:YES];
    [[self.loginController window] makeKeyAndOrderFront:self];
    [self.loginController showWindow:self];
}


- (IBAction)launchSubmit:(id)sender
{
    NSTask* task = [[NSTask alloc] init];
    task.launchPath = @"/usr/local/bin/p4vc";
    
    PLSOverseer* os = self.overseers[0];
    
    task.arguments  = @[@"-p", os.port, @"-u", os.user, @"-c", os.client, @"-d", os.path, @"-C", os.charset, @"submit"];

    [task launch];
}


// find the first connection needing a login, or just pass through the list quietly
- (IBAction)findConnectionNeedingLogin:(id)sender
{
    for(PLSOverseer* os in self.overseers)
    {
        if(os.needsLogin && !os.disabled)
        {
            [self showLogin:self overseer:os];
            return;
        }
    }
}


// set self as the Overseer delegate for any new Overseers
- (IBAction)connectionWindowClosed:(id)sender
{
    // reset our state icon appropriately
    BOOL needsLogin = false;
    BOOL connectionDown = false;
    
    for(PLSOverseer* os in self.overseers)
    {
        os.delegate = self;
        [os startTracking];
    
        // only include connections that aren't disabled
        if(!os.disabled)
        {
            needsLogin = needsLogin | os.needsLogin;
            connectionDown = connectionDown | os.connectionDown;
        }
    }
    
    
    if(needsLogin)
    {
        [self.statusItem setImage:[NSImage imageNamed:@"perforce_angry.png"]];
    }
    else if(connectionDown)
    {
        [self.statusItem setImage:[NSImage imageNamed:@"perforce_sad.png"]];
    }
    else
    {
        [self.statusItem setImage:[NSImage imageNamed:@"perforce.png"]];
    }
}


// notify the users we need to login
-(void)needPassword:(PLSOverseer*)os
{
    [self.statusItem setImage:[NSImage imageNamed:@"perforce_angry.png"]];
    [[self.statusMenu itemWithTitle:@"Login"] setEnabled:true];
    
    NSUserNotification *notification = [[NSUserNotification alloc] init];
    notification.title = @"Pulse requires you to login";
    notification.informativeText = @"Files will no longer be updated or submitted.";
    notification.soundName = NSUserNotificationDefaultSoundName;
    [[NSUserNotificationCenter defaultUserNotificationCenter] deliverNotification:notification];
}


// get the password right this minute
-(void)needPasswordNow:(PLSOverseer*)os
{
    // pop up the password dialog
    [self.statusItem setImage:[NSImage imageNamed:@"perforce_angry.png"]];
    [[self.statusMenu itemWithTitle:@"Login"] setEnabled:true];
    [self showLogin:self overseer:os];
}


// one of our Overseers lost contact with the Broodmother
-(void)connectionDown:(PLSOverseer*)os
{
    [self.statusItem setImage:[NSImage imageNamed:@"perforce_sad.png"]];
}


// check to see if all connections are up now
-(void)connectionUp:(PLSOverseer*)os
{
    BOOL disconnected = false;
    
    for(PLSOverseer* os in self.overseers)
    {
        if(!os.disabled)
        {
            disconnected = disconnected | os.connectionDown;
        }
    }
    
    if(!disconnected)
    {
        [self.statusItem setImage:[NSImage imageNamed:@"perforce.png"]];
    }
}


// one of our overseers has succesffuly logged in
-(void)loggedIn
{
    BOOL needsLogin = false;
    
    for(PLSOverseer* os in self.overseers)
    {
        if(!os.disabled)
        {
            needsLogin = needsLogin | os.needsLogin;
        }
    }
    
    if(!needsLogin)
    {
        [self.statusItem setImage:[NSImage imageNamed:@"perforce.png"]];
        [[self.statusMenu itemWithTitle:@"Login"] setEnabled:false];
    }
}

@end
